------------------------------------------------------------------------------
--		This file is ment to provide a layer of abstraction to buttons		--
--		FlexBar2 buttons do not use Action IDs, most code in this file		--
--		is aimed to emulate the behaviour of ActionButtons.					--
------------------------------------------------------------------------------

local Action =  FlexBar2:NewModule("Action", "AceEvent-2.0");
-- If this var is incremented, SetDefaults() will be called for this module thus resetting all its settings.
Action.Version = 2;

-- PickupAction function, this takes an action type (macro/item/spell) as first argument and its value (macro id/itemlink/spell name) as second
function Action:PickupAction(ActionType, ActionAttribute)
	-- Make sure the cursor is clear before picking anything up
    ClearCursor();
    FlexBar2:Debug("PickupAction:", ActionType, ActionAttribute);
    if(ActionAttribute) then
		-- Pickup the spell with its name
        if(ActionType == "spell") then
            PickupSpell(ActionAttribute);
		-- A bit more complicated, pickup an item using its cached BagID & BagSlot or using its cached InvSlot
        elseif(ActionType == "item" or ActionType == "itemlink") then
            local ItemInfo = FlexBar2:GetModule("ItemCache"):GetItemInfo(ActionAttribute);
            FlexBar2:Debug("Item: ",ActionAttribute);
            if(type(ItemInfo) == "table") then
                if(ItemInfo.Bag and ItemInfo.BagSlot) then
                    PickupContainerItem(ItemInfo.Bag, ItemInfo.BagSlot);
                elseif(ItemInfo.InvSlot) then
                    PickupInventoryItem(ItemInfo.InvSlot);
                else
                    FlexBar2:Print("Item currently not in the inventory, so can not pickup. Sorry :(");
                end                
            else
                FlexBar2:Print("Item currently not in the inventory, so can not pickup. Sorry :(");
            end
		-- Pickup the macro using its ID
        elseif(ActionType == "macro") then
            PickupMacro(ActionAttribute);
        end
    end
end

-- Declare the button mixin
Action.ButtonMixin = FlexBar2.OO.Mixin
{
    "LockAction",
    "UnlockAction",
    "Disable",
    "Enable",
	"ClearAction",
    "PickupCurrentAction",
	"PlaceAction",
	"GetActionTexture",
    "GetActionCooldown",
	"UpdateActionTexture",
    "UpdateCooldownSpinner",
    "PreClick_DisableButton",
    "PostClick_PickOldAction",
};

function Action.ButtonMixin:Activate()
	-- Register for dragging so action's can be dragged out
    self.Frame:RegisterForDrag("LeftButton");
	-- update the action texture when Shift Alt or Ctrl or pushed so that the texture can set itself to what its mapped to on that key
	self:RegisterEvent("MODIFIER_STATE_CHANGED", "UpdateActionTexture");
	-- Update the cooldown spinner when we get this event
    self:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN", "UpdateCooldownSpinner");
	-- Update button texture on state changes 
	self:RegisterScript("OnAttributeChanged", "UpdateActionTexture");
	-- A cache of old actions is kept, see later (at the PreClick_DisableButton function)
    self.OldActionCache = {};
	-- Update the cooldown text sub every 0.3 seconds
	self:RegisterEvent("FB2_UPDATE", function() self:UpdateTextSub("cd"); end);
	-- Update the itemcount text sub with the BAG_UPDATE event and the FB2_ActionChanged event
	self:RegisterEvent("BAG_UPDATE", function() self:UpdateTextSub("itemcount"); end);
	self:RegisterEvent("FB2_ActionChanged", function() self:UpdateTextSub("itemcount"); end);
end

-- ActionBUttons are locked by default
function Action.ButtonMixin:SetDefaults()
    self.InfoTable.LockAction = false;
end

-- Update old settings to new format
function Action.ButtonMixin:UpdateSettings(OldVersion)
	FlexBar2:Debug("Updating Action settings from " .. OldVersion .. " to " , Action.Version);
	-- Update 1->2: fixes bugs with conjured items by wiping the uniqueid of all items
	if(OldVersion <= 1) then
		local ItemInfo = FlexBar2:GetModule("ItemCache"):GetItemInfo(self:GetModifiedAttribute("item"));
		if(ItemInfo) then
			-- Found a bag id & a bag slot, get a new itemlink with the new uniqueid
			if(ItemInfo.Bag and ItemInfo.BagSlot) then
				-- Strip out uniqueid from the itemlinks to fix nasty bugs
				local ItemLink = GetContainerItemLink(ItemInfo.Bag, ItemInfo.BagSlot);
				ItemLink = string.gsub(ItemLink, "(.*):(.-)$", "%1:0");
				self:SetModifiedAttribute("item", "itemlink:" .. ItemLink);
			-- Found an equiped item, get a new itemlink with the new uniqueid
			elseif(ItemInfo.InvSlot) then
				-- Strip out uniqueid from the itemlinks to fix nasty bugs
				local ItemLink = GetInventoryItemLink(ItemInfo.InvSlot)
				ItemLink = string.gsub(ItemLink, "(.*):(.-)$", "%1:0");
				self:SetModifiedAttribute("item", "itemlink:" .. ItemLink);
			end
		end
	end
	return true;
end
-- Load the lock setting & apply it
function Action.ButtonMixin:LoadSettings()
    if(self.InfoTable.LockAction == true) then
        self:LockAction();
    else
        self:UnlockAction();
    end
end

-- Load the button's texture and set its cooldown spinner upon load
function Action.ButtonMixin:Initialize()
    self:UpdateActionTexture();
    self:UpdateCooldownSpinner();
end

-- Unregister all drag scripts when the button is locked
function Action.ButtonMixin:LockAction()
	-- Disable regular dragging
    self:UnregisterScript("OnDragStart", "PickupCurrentAction");
    self:UnregisterScript("OnReceiveDrag", "PlaceAction");
    self:UnregisterScript("PreClick", "PreClick_DisableButton");
    self:UnregisterScript("PostClick", "PostClick_PickOldAction");
end

-- Register all drag scripts when the button is unlocked
function Action.ButtonMixin:UnlockAction()
	-- Enable regular dragging
    self:RegisterScript("OnDragStart", "PickupCurrentAction");
    self:RegisterScript("OnReceiveDrag", "PlaceAction");
    self:RegisterScript("PreClick", "PreClick_DisableButton");
    self:RegisterScript("PostClick", "PostClick_PickOldAction");	
end

-- Disable the button, this should not be used for log disablings. only works OUT of combat
function Action.ButtonMixin:Disable()
    if(InCombatLockdown() == nil) then
        self.DisabledActionType = self:GetModifiedAttribute("type");
        self:SetModifiedAttribute("type", nil, false);
    end
end

-- Reenable the button. only works OUT of combat
function Action.ButtonMixin:Enable()
    if(InCombatLockdown() == nil and self.DisabledActionType ~= nil) then
        self:SetModifiedAttribute("type", self.DisabledActionType, false);
        self.DisabledActionType = nil;
    end
end

-- Get the texture that's suposed to currently be on the button
function Action.ButtonMixin:GetActionTexture()
	local ActionType = self:GetModifiedAttribute("type");
	-- If its a spell, get the spell's texture using its name
	if(ActionType == "spell") then
		return GetSpellTexture(self:GetModifiedAttribute("spell") or "") or "";
	-- If its an item get its texture with GetItemInfo using its itemlink
	elseif(ActionType == "item") then
		local ItemName, ItemLink, ItemRarity, ItemLevel, ItemMinLevel, ItemType, ItemSubType, ItemStackCount, ItemEquipLoc, ItemTexture = GetItemInfo(self:GetModifiedAttribute("itemlink") or "");
		return ItemTexture or "";
	-- If its a macro get the texture using the macro ID
	elseif(ActionType == "macro") then
		local MacroName, MacroTexture, MacroBody, MacroLocal = GetMacroInfo(self:GetModifiedAttribute("macro"));
		return MacroTexture or "";
	else
		return "";
	end
end

-- set the actionbutton's texture using GetActionTexture
function Action.ButtonMixin:UpdateActionTexture()
	self:SetIcon(self:GetActionTexture());
end

-- Get an actionbutton's cooldown
function Action.ButtonMixin:GetActionCooldown()
    local ActionType = self:GetModifiedAttribute("type");
	-- If its a spell use GetSpellCooldown with the spell's name
    if(ActionType == "spell") then
        return GetSpellCooldown(self:GetModifiedAttribute("spell"));
	-- If its an item either use GetContainerItemCooldown or GetInventoryItemCooldown depending on wether the item is equiped or not
    elseif(ActionType == "item") then
        local ItemInfo = FlexBar2:GetModule("ItemCache"):GetItemInfo(self:GetModifiedAttribute("item"));
        if(ItemInfo) then
            if(ItemInfo.Bag and ItemInfo.BagSlot) then
                return GetContainerItemCooldown(ItemInfo.Bag, ItemInfo.BagSlot);
            elseif(ItemInfo.InvSlot) then
                return GetInventoryItemCooldown("player", ItemInfo.InvSlot);
            end
        end
    elseif(ActionType == "macro") then
        -- Unsupported as of yet
    end
end

-- Update the button's cooldown spinner using GetActionCooldown's return stuff
function Action.ButtonMixin:UpdateCooldownSpinner() 
    local CooldownFrame = self.Frame.Cooldown;
    local start, duration, enable = self:GetActionCooldown();
    FlexBar2:Debug("Updating cooldown spinner: ", start, duration, enable);
    if(start ~= nil and duration ~= nil and enable ~= nil and start > 0 and duration > 0 and enable > 0) then
        CooldownFrame:SetCooldown(start, duration);
        CooldownFrame:Show();
    else
        CooldownFrame:Hide();
    end
end

-- Clear the button's action for this state only. only works OUT of combat
function Action.ButtonMixin:ClearAction()
    if(InCombatLockdown() == nil) then
        self:SetModifiedAttribute("type", nil);
        self:SetModifiedAttribute("spell", nil);
        self:SetModifiedAttribute("macro", nil);
        self:SetModifiedAttribute("item", nil);
		-- Custom attributes for extra info
        self:SetModifiedAttribute("itemlink", nil);
		-- Update the textute after nil'ing out all this stuff
        self:UpdateActionTexture();
    end
end

-- Disable the button on preclick if the cursor contains anything, this allows us to put the action on the button without triggering a click. only works OUT of combat
function Action.ButtonMixin:PreClick_DisableButton()
    if(GetCursorInfo() ~= nil and InCombatLockdown() == nil) then
        local OldActionCache = self.OldActionCache;
        OldActionCache.ActionType = self:GetModifiedAttribute("type");
        OldActionCache.ActionAttribute = self:GetModifiedAttribute(OldActionCache.ActionType);
		-- Place any actions currently on the cursor on the button
        self:PlaceAction();
		-- Disable the button so it wont catch the click and trigger the newly placed action
        self:Disable();
    end
end

-- Reenable the button on postclick so it can catch clicks again
function Action.ButtonMixin:PostClick_PickOldAction()
    local OldActionCache = self.OldActionCache;
    if(OldActionCache.ActionType ~= nil and OldActionCache.ActionType ~= "" and OldActionCache.ActionAttribute ~= nil and OldActionCache.ActionAttribute ~= "") then
        Action:PickupAction(OldActionCache.ActionType, OldActionCache.ActionAttribute);
        OldActionCache.ActionType = nil;
        OldActionCache.ActionAttribute = nil;
    end
	-- Reenable the button so it can catch clicks again
    self:Enable();
end

-- Pickup the current action from the button to the cursor; only works OUT of combat
function Action.ButtonMixin:PickupCurrentAction()
	if(InCombatLockdown() == nil) then
		-- Get Action's type
        local ActionType = self:GetModifiedAttribute("type");
		-- Pickup the action using its type and the type's value
        Action:PickupAction(ActionType, self:GetModifiedAttribute(ActionType));
		-- Clear any actions currently on the cursor
        self:ClearAction();
		-- Update cooldown spinner & texture
        self:UpdateCooldownSpinner();
        self:UpdateActionTexture();
		-- Trigger an event for this, so other modules can behave accordingly, pass self as a paremeter so a check can be made to only update this button
        self:TriggerEvent("FB2_ActionChanged", self);
	end
end

-- Place the action currently on the cursor to the button. only works OUT of combat
function Action.ButtonMixin:PlaceAction()
    if(InCombatLockdown() == nil) then
		-- Get info from the cursor, does not return anything for pets, so unsupported
        local ActionType, ActionId, ActionBook = GetCursorInfo()
		FlexBar2:Debug("PlaceAction:", ActionType, ActionId, ActionBook);
		-- If there is currently an action on the button, pick it up
        if(ActionType ~= nil) then self:PickupCurrentAction(); end
		-- ActionType is a spell, get the spell's name & rank, format it into a SpellName(Rank X) format and save it
        if(ActionType == "spell") then
            self:SetModifiedAttribute("type", "spell");
            local SpellName, SpellRank = GetSpellName(ActionId, ActionBook);
			-- Parse a rank number out of Rank X
			_, _, SpellRank = string.find(SpellRank, "^Rank (%d+)$");
			SpellRank = tonumber(SpellRank)
			-- If the spell is the highest rank available to the player, dont specify the rank, that way it'll autorank up when the user learns a new rank
            if(type(SpellRank) == "number") then
				-- Find out if the spell is the highest rank by requesting the texture of the spell one rank higher, if this isn't nil, then a rank above exists
				local HigherRankExists = GetSpellTexture(SpellName .. "(Rank " .. SpellRank + 1 .. ")") ~= nil;
                if(HigherRankExists) then
					SpellName = SpellName .. "(Rank " .. SpellRank .. ")";
				end
            end			
            self:SetModifiedAttribute("spell", SpellName);
		-- Get the itemlink & save it. the itemlink attribute is not needed but is kept so to not have to remove itemlink: from the front of the item attribute when the raw item link is needed (blizzard funcs :p)
        elseif(ActionType == "item") then
			self:SetModifiedAttribute("type", "item");
			local ItemLink = string.gsub(ItemLink, "(.*):(.-)$", "%1:0");
            self:SetModifiedAttribute("item", "itemlink:" .. ItemLink);
            self:SetModifiedAttribute("itemlink", ItemLink);
		-- Macro, just save the macro's ID
        elseif(ActionType == "macro") then
			self:SetModifiedAttribute("type", "macro");
			self:SetModifiedAttribute("macro", ActionId);
        end
		-- Update cooldown spinner & button texture
        self:UpdateCooldownSpinner();
		self:UpdateActionTexture();
		-- Trigger an event for this, so other modules can behave accordingly, pass self as a paremeter so a check can be made to only update this button
        self:TriggerEvent("FB2_ActionChanged", self);
    else
        FlexBar2:Print("Actions cannot be changed in combat, sorry");
    end
end

-- Text subs --
local TextSubs = FlexBar2:GetModule("TextSubs");

TextSubs:RegisterTextSub("cd", function(Button)
	-- Get the cooldown currently on the button
	local StartTime, Duration, Enable = Button:GetActionCooldown();
	if(StartTime and StartTime > 0 and Duration > 0 and Enable > 0) then
		local TimeLeft = floor(Duration - GetTime() + StartTime + 0.5);
		-- Timer smaller than 0 or 0
		if(TimeLeft <= 0 or Duration <= 1.5) then			
			return "";
		-- Timer smaller than 1 min
		elseif(TimeLeft < 60) then
			return TimeLeft .. "s";
		-- Timer between 1 to 10 mins
		elseif(TimeLeft < 600) then
			local SecondsLeft = tostring((TimeLeft % 60));
			if(#SecondsLeft == 1) then
				SecondsLeft = "0" .. SecondsLeft;
			end
			return floor(TimeLeft/60) .. ":" .. SecondsLeft;
		-- Timer bigger then 10 mins
		else
			return floor(TimeLeft/60) .. "m";
		end
	end
	return "";
end);

TextSubs:RegisterTextSub("itemcount", function(Button, Item)
	-- No item given, find out what item is on this button instead
	if(Item == "") then
		local ItemLink = Button:GetModifiedAttribute("itemlink");
		if(ItemLink) then
			return GetItemCount(ItemLink or "") or "";
		end
	-- Item was given, ez-mode
	else
		return GetItemCount(Item or "") or "";
	end
	-- Fallback
	return "";
end);

-- GUI --
local Waterfall = FlexBar2.Waterfall;
local Status_LockAction;
function Action.ButtonWaterfall(Action, ButtonName, ButtonTable)
	if(Action == "edit") then
		-- Retrieve those settings when the GUI gets shown
        Status_LockAction = FlexBar2:GetSetting(ButtonTable, "LockAction");
		-- Lock Action checkbox --
		Waterfall:AddControl(
		"type","checkbox",
		"text","Lock Action",
		"width",300,
		"causeRefresh", false,
		"getFunc" ,function() return Status_LockAction; end,
		"setFunc", function(LockAction) Status_LockAction = LockAction; end);
	elseif(Action == "save") then
		local ButtonList = FlexBar2.Buttons;
		FlexBar2:Debug(Status_LockAction);
		-- Save lock settings & apply em
        FlexBar2:ApplySetting(ButtonTable, "LockAction", Status_LockAction);
        if(Status_LockAction == true) then
            ButtonList[ButtonTable]:LockAction();
        else
            ButtonList[ButtonTable]:UnlockAction();
        end
	end
end

-- SlashCommands --

-- /fb lock command
local function SlashCommand_LockAction(Params)
	local ButtonTable = FlexBar2:ToButtonTable(Params.button);
	FlexBar2:ApplySetting(ButtonTable, "LockAction", true);
	local ButtonList = FlexBar2.Buttons;
	for _, ButtonName in ipairs(ButtonTable) do
		ButtonList[ButtonName]:LockAction();
	end
end

-- /fb unlock command
local function SlashCommand_UnlockAction(Params)
	local ButtonTable = FlexBar2:ToButtonTable(Params.button);
	FlexBar2:ApplySetting(ButtonTable, "LockAction", false);
	local ButtonList = FlexBar2.Buttons;
	for _, ButtonName in ipairs(ButtonTable) do
		ButtonList[ButtonName]:UnlockAction();
	end
end

local SlashCommands = FlexBar2.SlashCommands;
-- Localize this neat verify function for faster access
local VerifyExistance = FlexBar2.SlashCommands.VerifyExistance;
-- AddCommand format :AddCommand(Command, Function, Group, Help)
-- AddParam format :AddParam(Command, Param, Required, AllowedValues, DefaultValue, VerifyFunc) 
-- everything here should be self explainative knowing the syntax
SlashCommands:AddCommand("lockaction", SlashCommand_LockAction, "action", "/FlexBar2 lockaction button=<Button>");
SlashCommands:SetAlias("lockaction", "actionlock");
	SlashCommands:AddParam("lockaction", "button", true, nil, nil, VerifyExistance);
SlashCommands:AddCommand("unlockaction", SlashCommand_UnlockAction, "action", "/FlexBar2 unlockaction button=<Button>");
SlashCommands:SetAlias("unlockaction", "actionunlock");
	SlashCommands:AddParam("unlockaction", "button", true, nil, nil, VerifyExistance);
